## Function to calculate weights for a testpoint
ComputeGaussianKernel = function(x,y,lambda){
  x = as.numeric(x)
  y = as.numeric(y)
  lambda = as.numeric(lambda)
  kernel = (1/(sqrt(2*pi)*lambda))*exp(-(x-y)**2/(2*lambda**2))
  return(kernel)
}

ComputeVonMisesKernel = function(D,D0, nu){
  D = as.numeric(D)
  D0 = as.numeric(D0)
  nu = as.numeric(nu)
  kernel =  exp(nu*cos(D-D0))/(2*pi*besselI(x=nu,nu=0))
  return(kernel)
}


## Function to calculate weights for a trivariate kernel given the bivariate kernel and the third input variable
calculateWeights = function(train.x,testpoint,bandwidth.x,id.spd,id.dir){
  if (ncol(train.x) == 1){
    weights = matrix(NA,nrow(train.x),1)
    kernel = ComputeGaussianKernel(train.x[,id.spd],testpoint[id.spd],bandwidth.x[id.spd])
    weights[,1] = kernel/sum(kernel)
  } else if (ncol(train.x)==2){
    weights = matrix(NA,nrow(train.x),1)
    kernel =ComputeGaussianKernel(train.x[,id.spd],testpoint[id.spd],bandwidth.x[id.spd])*ComputeVonMisesKernel(train.x[,id.dir],testpoint[id.dir],bandwidth.x[id.dir])
    weights[,1] = kernel/sum(kernel)

  } else {
    weights = matrix(NA,nrow(train.x),ncol(train.x)-2)
    inputCols = c(1:ncol(train.x))
    id.third = setdiff(inputCols,c(id.spd,id.dir))
    for (i in 1:length(id.third)){
      kernel = ComputeGaussianKernel(train.x[,id.spd],testpoint[id.spd],bandwidth.x[id.spd])*ComputeVonMisesKernel(train.x[,id.dir],testpoint[id.dir],bandwidth.x[id.dir])*ComputeGaussianKernel(train.x[,id.third[i]],testpoint[id.third[i]],bandwidth.x[id.third[i]])
      weights[,i] = kernel/sum(kernel)
    }
  }
  return(weights)
}

## Function to get bandwidth of x using dpi
computeBandwidthX = function(train.y,train.x,id.dir = 2, units.dir = "degrees"){
  train.x = as.matrix(train.x)
  train.y = as.numeric(train.y)
  library(KernSmooth)
  bandwidth = NULL
  for (i in 1:ncol(train.x)){
    bandwidth[i] = dpill(train.x[,i],train.y)
  }
  if (units.dir == "degrees"){
    bandwidth[id.dir] = bandwidth[id.dir]*pi/180
  }
  bandwidth[id.dir] = 1/((bandwidth[id.dir])^2)
  return(bandwidth)
}

plusSign = function(x){
  if (sign(x) == 1){return(1)} else {return(0)}
}
## Function to compute CRPS Integrand
CRPSIntegrand = function(y,test.y,train.y,bandwidth.y,weights){
  integrand = rep(NA,length(y))
  for (i in 1:length(y)){
    integrand[i] = ((rowSums(weights)%*%pnorm(y[i],train.y,bandwidth.y))/ncol(weights) - plusSign(y[i]-test.y))**2
  }
  return(as.numeric(integrand))
}

KernelCDF = function(y,train.y,bandwidth.y,weights){
  cdf = rep(NA,length(y))
  for (i in 1:length(y)){
    cdf[i] = (rowSums(weights)%*%pnorm(y[i],train.y,bandwidth.y))/ncol(weights)
  }
  return(as.numeric(cdf))
}


## Function to compute CRPS using trapz
ComputeCRPS = function(test.y, test.x,train.x, train.y,bandwidth.y, bandwidth.x,id.spd=1,id.dir=2,units.dir="degrees"){
  library(caTools)
  if (units.dir == "degrees"){
    train.x[,id.dir] = train.x[,id.dir]*pi/180
    test.x[,id.dir] = test.x[,id.dir]*pi/180
  }
  Integral = rep(NA,nrow(test.x))
  for (i in 1:nrow(test.x)){
    weights = calculateWeights(train.x,test.x[i,],bandwidth.x,id.spd,id.dir)
    x = seq(0,100,1)
    y = CRPSIntegrand(x,test.y[i],train.y,bandwidth.y,weights)
    Integral[i] = trapz(x,y)
  }
  return(mean(Integral))
}



## Function to compute bart CRPS
ComputeBartCRPS = function(yhat.test,y.test){
  library(caTools)
 Integral = rep(NA,length(y.test))
  for (i in 1:length(Integral)){
   dy = seq(0,100,0.1)
   y_dist = yhat.test[,i]
   Integrand = ComputeBartCRPSIntegrand(dy,y_dist,y.test[i])
   Integral[i] = trapz(dy,Integrand)
  } 
 return(mean(Integral))
}

## Function to compute cdf for bart from posterior samples
ComputeBartCRPSIntegrand = function(y,y_dist,y.test){
  Integrand = rep(NA,length(y))
  for (i in 1:length(Integrand)){
    Integrand[i] = (length(which(y_dist<=y[i]))/length(y_dist) - plusSign(y[i] - y.test))**2
  }
  return(Integrand)
}

## Function to compute Conditional density of y $f(y|x)$
computeCondDensity = function(test.y,train.y,bandwidth.y,weights){
  density = rep(NA, length(test.y))
  for (i in 1:length(test.y)){ 
    trainDensity = dnorm(test.y[i],mean = train.y,sd = bandwidth.y)
    density[i] = (rowSums(weights)%*%trainDensity)/ncol(weights)
  }
  return(as.vector(density))
}


## Function to compute optimal bandwidth of Y based on the greedy algorithm.
computeBandwidthY = function(y,x,bandwidth.x,id.spd=1,id.dir=2,units.dir="degrees",accuracy = 0.1,lower = 0, upper = 5){
  library(caTools)
  solution = optimize(f = computeISE, lower = lower, upper = upper, tol = accuracy, bandwidth.x=bandwidth.x, x = x, y = y,id.spd = id.spd, id.dir = id.dir,units.dir=units.dir)
  return(as.numeric(solution$minimum))
}


## Function to compute the first two parts of Integrated squared loss $I1-2I2$
computeISE = function(bandwidth.y,bandwidth.x,y,x,id.spd,id.dir,units.dir){
  y = as.numeric(y)
  x = as.matrix(x)
  if (units.dir == "degrees"){
    x[,id.dir] = x[,id.dir]*pi/180
  }
  I1 = 0
  I2 = 0
  for (i in 1:length(y)){
    train.x = x[-i,]
    train.y = y[-i]
    test.x = x[i,]
    test.y = y[i]
    BivariateKernel = ComputeGaussianKernel(train.x[,id.spd],test.x[id.spd],bandwidth.x[id.spd])*ComputeVonMisesKernel(train.x[,id.dir],test.x[id.dir],bandwidth.x[id.dir])
    weights = as.matrix(BivariateKernel/sum(BivariateKernel))
    I1 = I1 + computeI1(train.y, bandwidth.y,weights)
    I2 = I2 + computeCondDensity(test.y,train.y,bandwidth.y,weights)
  }
  ISE = (I1 - 2*I2)/length(y)
  return(ISE)
}

## Function to compute I1 for one training point
computeI1 = function(train.y,bandwidth.y,weights){
  library(caTools)
  x = seq(0,100,1)
  y = squaredCondDensity(x,train.y = train.y, bandwidth.y = bandwidth.y,weights = weights)
  Integral = trapz(x,y)
  return(Integral)
}

## Function to compute squared conditional density for a testpoint (to be used in calculation of I1)
squaredCondDensity = function(test.y,train.y,bandwidth.y,weights){
  density = rep(NA, length(test.y))
  for (i in 1:length(test.y)){ 
    trainDensity = dnorm(train.y,mean = test.y[i],sd = bandwidth.y)
    density[i] = ((rowSums(weights)%*%trainDensity)/ncol(weights))**2
  }
  return(as.vector(density))
}

## Function to predict response from IEC binning method
binning = function(train.y, train.x, bin_width, test.x){
  library(stats)
  train.y = as.numeric(train.y)
  train.x = as.numeric(train.x)
  test.x = as.numeric(test.x)
  start = round(min(train.x))
  end = round(max(train.x))
  n_bins = round((end - start)/bin_width,0)
  x_bin = NULL
  y_bin = NULL
  for (n in 1:n_bins){
    bin_element = which(train.x>(start+(n-1)*bin_width) & train.x<(start+n*bin_width))
    x_bin[n] = start + ((2*n-1)*bin_width)/2
    y_bin[n] = mean(train.y[bin_element])
  }
  binned_data = data.frame(x_bin, y_bin)
  binned_data = binned_data[which(is.na(binned_data$y_bin)==F),]
  splinefit = smooth.spline(x = binned_data$x_bin, y= binned_data$y_bin , all.knots = T)
  y_pred = predict(splinefit,test.x)$y
  return(y_pred)
}


### Function to get mean estimate
kernpred = function(train.y, train.x, test.x, id.spd = 1, id.dir = 2, units.dir = "degrees"){
  train.x = as.matrix(train.x)
  test.x = as.matrix(test.x)
  bandwidth.x = computeBandwidthX(train.y,train.x)
  if (ncol(train.x)>1){
    if(units.dir == "degrees"){
      train.x[,id.dir] = train.x[,id.dir]*pi/180
      test.x[,id.dir] = test.x[,id.dir]*pi/180
      bandwidth.x[id.dir] = bandwidth.x[id.dir]*pi/180
    }
    bandwidth.x[id.dir] = 1/((bandwidth.x[id.dir])^2)
  }
  pred = rep(NA, nrow(test.x))
  for (i in 1:length(pred)){
    weights = calculateWeights(train.x,test.x[i,],bandwidth.x,id.spd,id.dir)
    pred[i] = (rowSums(weights)%*%train.y)/ncol(weights)
  }
  return(pred)
}


## Function to get optimal 'k' for kNN based on cross validation
computeOptimalK = function(data.x, data.y, CV.folds, range.k){
  library(FNN)
  best.k = NULL
  best.err = Inf
  data.x = as.matrix(data.x)
  data.y = as.matrix(data.y)
  cv_sample = sample(nrow(data.x))
  folds = cut(seq(1:length(cv_sample)),breaks = CV.folds, labels = F)
  for (i in 1:length(range.k)){
    CV_Err = rep(NA,CV.folds)
    for (f in 1:CV.folds){
      testindex = cv_sample[which(folds == f,arr.ind = T)]
      train.x = data.x[-testindex,]
      train.y = data.y[-testindex,]
      test.x = data.x[testindex,]
      test.y = data.y[testindex,]
      pred.y = knn.reg(y = train.y, train = train.x, test = test.x, k = range.k[i])$pred
      CV_Err[f] = sqrt(mean((test.y-pred.y)^2))
    }
    if (mean(CV_Err)< best.err){
      best.err = mean(CV_Err)
      best.k = range.k[i]
    }
  }
  return(best.k)
}